#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os

import pytest
import pytest_asyncio
from mock import mock

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data
import octobot_trading.errors as errors
from octobot_trading.personal_data import SellLimitOrder, BuyLimitOrder, BuyMarketOrder, SellMarketOrder

from tests import event_loop
from tests.exchanges import future_simulated_exchange_manager
from tests.exchanges.traders import future_trader_simulator_with_default_linear, \
    future_trader_simulator_with_default_inverse, DEFAULT_FUTURE_SYMBOL
from tests.test_utils.random_numbers import decimal_random_price, decimal_random_quantity

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture
async def btc_usdt_future_trader_simulator_with_default_linear(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    position_inst = personal_data.LinearPosition(trader_inst, default_contract)
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    # force unknown position to update side later on when updating size
    position_inst.side = enums.PositionSide.UNKNOWN
    return config, exchange_manager_inst, trader_inst, default_contract, position_inst


@pytest_asyncio.fixture
async def btc_usdt_future_trader_simulator_with_default_inverse(future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    position_inst = personal_data.InversePosition(trader_inst, default_contract)
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    position_inst.side = enums.PositionSide.UNKNOWN
    return config, exchange_manager_inst, trader_inst, default_contract, position_inst


async def test_update_entry_price(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst \
        = btc_usdt_future_trader_simulator_with_default_linear

    assert position_inst.entry_price == constants.ZERO
    assert position_inst.mark_price == constants.ZERO

    mark_price = decimal_random_price(constants.ONE, decimal.Decimal(1000))
    await position_inst.update(mark_price=mark_price, update_size=constants.ONE)
    assert position_inst.entry_price == mark_price
    assert position_inst.mark_price == mark_price


async def test_update_margin_linear(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst \
        = btc_usdt_future_trader_simulator_with_default_linear

    default_contract.set_current_leverage(decimal.Decimal(10))

    assert position_inst.entry_price == constants.ZERO
    assert position_inst.mark_price == constants.ZERO
    assert position_inst.initial_margin == constants.ZERO
    assert position_inst.size == constants.ZERO
    assert position_inst.quantity == constants.ZERO
    assert position_inst.creation_time > 0
    initial_creation_time = 123
    # force changing creation time
    position_inst.creation_time = 123

    await position_inst.update(mark_price=constants.ONE, update_margin=decimal.Decimal(5))
    assert position_inst.mark_price == constants.ONE
    assert position_inst.initial_margin == decimal.Decimal(5)
    assert position_inst.entry_price == constants.ONE
    assert position_inst.margin == decimal.Decimal("5.0300")    # initial margin + closing fees
    assert position_inst.size == decimal.Decimal(50)
    # creation time got updated: site changed
    assert initial_creation_time != position_inst.creation_time
    updated_creation_time = position_inst.creation_time

    await position_inst.update(mark_price=decimal.Decimal("1.1"), update_margin=-decimal.Decimal(3))
    assert position_inst.mark_price == decimal.Decimal("1.1")
    assert position_inst.initial_margin == decimal.Decimal(2)
    assert position_inst.entry_price == constants.ONE   # did not change
    assert position_inst.margin == decimal.Decimal("2.01320")
    assert position_inst.size == decimal.Decimal(20)
    # creation time did not get updated
    assert updated_creation_time == position_inst.creation_time


async def test_update_margin_inverse(btc_usdt_future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_inverse

    default_contract.set_current_leverage(decimal.Decimal(5))

    assert position_inst.entry_price == constants.ZERO
    assert position_inst.mark_price == constants.ZERO
    assert position_inst.initial_margin == constants.ZERO
    assert position_inst.size == constants.ZERO

    await position_inst.update(mark_price=constants.ONE, update_margin=decimal.Decimal(8) / constants.ONE_HUNDRED)
    assert position_inst.mark_price == constants.ONE
    assert position_inst.entry_price == constants.ONE
    assert position_inst.initial_margin == decimal.Decimal('0.08')
    assert position_inst.margin == decimal.Decimal("0.080288")  # initial margin + closing fees
    assert position_inst.size == decimal.Decimal('0.4')

    await position_inst.update(mark_price=constants.ONE, update_margin=-constants.ONE / constants.ONE_HUNDRED)
    assert position_inst.mark_price == constants.ONE
    assert position_inst.entry_price == constants.ONE   # did not change
    assert position_inst.initial_margin == decimal.Decimal('0.07')  # - 0.01
    assert position_inst.margin == decimal.Decimal("0.070252")
    assert position_inst.size == decimal.Decimal('0.35')

    await position_inst.update(mark_price=decimal.Decimal("0.9"), update_margin=-constants.ONE / constants.ONE_HUNDRED)
    assert position_inst.mark_price == decimal.Decimal("0.9")
    assert position_inst.entry_price == constants.ONE   # did not change
    assert position_inst.initial_margin == decimal.Decimal('0.06')  # - 0.01
    assert position_inst.margin == decimal.Decimal("0.06024")
    assert position_inst.size == decimal.Decimal('0.3')


async def test_update_entry_price_when_switching_side_on_one_way(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.initialize()
    position_inst.update_from_raw(
        {
            enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL
        }
    )

    mark_price = decimal.Decimal(80)
    await position_inst.update(mark_price=mark_price, update_size=decimal.Decimal(0.01))
    assert position_inst.entry_price == mark_price
    assert position_inst.mark_price == mark_price

    # updating mark price
    await position_inst.update(mark_price=decimal.Decimal(90))
    assert position_inst.is_long()

    # switching long to short with new mark price
    mark_price = decimal.Decimal(100)
    await position_inst.update(mark_price=mark_price, update_size=-decimal.Decimal(0.02))
    assert position_inst.is_short()
    assert position_inst.mark_price == mark_price
    assert position_inst.entry_price == mark_price

    # updating mark price
    await position_inst.update(mark_price=decimal.Decimal(60))
    assert position_inst.is_short()

    # switching short to long with new mark price
    mark_price = decimal.Decimal(50)
    await position_inst.update(mark_price=mark_price, update_size=decimal.Decimal(0.03))
    assert position_inst.is_long()
    assert position_inst.mark_price == mark_price
    assert position_inst.entry_price == mark_price


async def test_update_update_quantity(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    assert position_inst.quantity == constants.ZERO

    quantity = decimal_random_quantity(1)
    await position_inst.update(update_size=quantity)
    assert position_inst.quantity == quantity


async def test_invalid_update(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    portfolio = exchange_manager_inst.exchange_personal_data.portfolio_manager.portfolio.get_currency_portfolio("BTC")

    assert position_inst.quantity == constants.ZERO

    async def _ensure_no_position_change(mark_price, update_size):
        with pytest.raises(errors.PortfolioNegativeValueError):
            await position_inst.update(mark_price=mark_price, update_size=update_size)
        # Did not affect position or portfolio data
        assert position_inst.quantity == position_inst.entry_price == position_inst.mark_price == position_inst.size == \
               constants.ZERO
        assert portfolio.available == decimal.Decimal('10')
        assert portfolio.total == decimal.Decimal('10')
        assert portfolio.order_margin == constants.ZERO
        assert portfolio.position_margin == constants.ZERO
        assert portfolio.unrealized_pnl == constants.ZERO

    await _ensure_no_position_change(decimal.Decimal(50), decimal.Decimal(100000))
    await _ensure_no_position_change(decimal.Decimal(50), decimal.Decimal(-100000))


async def test__check_and_update_size_with_one_way_position_mode(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)

    if not os.getenv('CYTHON_IGNORE'):
        # LONG
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(10))
        position_inst._check_and_update_size(decimal.Decimal(10))
        assert position_inst.size == decimal.Decimal(20)
        position_inst._check_and_update_size(decimal.Decimal(-10))
        assert position_inst.size == decimal.Decimal(10)
        position_inst._check_and_update_size(decimal.Decimal(-30))
        assert position_inst.size == decimal.Decimal(-20)

        # SHORT
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(-100))
        position_inst._check_and_update_size(decimal.Decimal(-1.5))
        assert position_inst.size == decimal.Decimal(-101.5)
        position_inst._check_and_update_size(decimal.Decimal(51.5))
        assert position_inst.size == decimal.Decimal(-50)
        position_inst._check_and_update_size(decimal.Decimal(100))
        assert position_inst.size == decimal.Decimal(50)


async def test__check_and_update_size_with_hedge_position_mode(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)

    if not os.getenv('CYTHON_IGNORE'):
        # LONG
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(100))
        position_inst._check_and_update_size(decimal.Decimal(-5))
        assert position_inst.size == decimal.Decimal(95)
        position_inst._check_and_update_size(decimal.Decimal("-66.481231232156215215874878"))
        assert position_inst.size == decimal.Decimal("28.518768767843784784125122")
        position_inst._check_and_update_size(decimal.Decimal(-450))
        assert position_inst.size == constants.ZERO  # position should be closed

        # SHORT
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(-10))
        position_inst._check_and_update_size(decimal.Decimal(-10))
        assert position_inst.size == decimal.Decimal(-20)
        position_inst._check_and_update_size(decimal.Decimal(10))
        assert position_inst.size == decimal.Decimal(-10)
        position_inst._check_and_update_size(decimal.Decimal(50))
        assert position_inst.size == constants.ZERO  # position should be closed


async def test__is_update_increasing_size(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)

    if not os.getenv('CYTHON_IGNORE'):
        # Closed
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        assert position_inst._is_update_increasing_size(decimal.Decimal(-5))
        assert position_inst._is_update_increasing_size(decimal.Decimal(5))

        # LONG
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(100))
        assert not position_inst._is_update_increasing_size(decimal.Decimal(-5))
        assert position_inst._is_update_increasing_size(decimal.Decimal(5))

        # SHORT
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(-100))
        assert position_inst._is_update_increasing_size(decimal.Decimal(-5))
        assert not position_inst._is_update_increasing_size(decimal.Decimal(5))


async def test__is_update_closing(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)

    if not os.getenv('CYTHON_IGNORE'):
        # Idle
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        assert position_inst._is_update_closing(decimal.Decimal(-5))
        assert position_inst._is_update_closing(decimal.Decimal(5))

        # LONG
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(100))
        assert not position_inst._is_update_closing(decimal.Decimal(-5))
        assert not position_inst._is_update_closing(decimal.Decimal(5))
        assert not position_inst._is_update_closing(decimal.Decimal(100))
        assert not position_inst._is_update_closing(decimal.Decimal(500))
        assert not position_inst._is_update_closing(decimal.Decimal(99))
        assert position_inst._is_update_closing(decimal.Decimal(-100))
        assert not position_inst._is_update_closing(decimal.Decimal(-99))
        assert position_inst._is_update_closing(decimal.Decimal(-500))

        # SHORT
        position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
        await position_inst.update(update_size=decimal.Decimal(-100))
        assert not position_inst._is_update_closing(decimal.Decimal(-5))
        assert not position_inst._is_update_closing(decimal.Decimal(5))
        assert position_inst._is_update_closing(decimal.Decimal(100))
        assert position_inst._is_update_closing(decimal.Decimal(500))
        assert not position_inst._is_update_closing(decimal.Decimal(99))
        assert not position_inst._is_update_closing(decimal.Decimal(-100))
        assert not position_inst._is_update_closing(decimal.Decimal(-99))
        assert not position_inst._is_update_closing(decimal.Decimal(-500))


async def test_get_quantity_to_close(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    position_inst = personal_data.LinearPosition(trader_inst, default_contract)
    quantity = decimal_random_quantity(1)
    await position_inst.update(update_size=quantity)
    assert position_inst.get_quantity_to_close() == -quantity

    position_inst = personal_data.LinearPosition(trader_inst, default_contract)
    quantity = -decimal_random_quantity(1)
    await position_inst.update(update_size=quantity)
    assert position_inst.get_quantity_to_close() == -quantity


async def test_update_size_from_order_with_long_one_way_position(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=constants.ONE_HUNDRED)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(2),
                      price=decimal.Decimal(20))
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == decimal.Decimal(98)
    assert position_inst.is_long()


async def test_update_size_from_reversing_order_with_long_one_way_position(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=constants.ONE_HUNDRED)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=constants.ONE_HUNDRED + decimal.Decimal(50), # selling 150 => end size is -50
                      price=decimal.Decimal(20),
                      filled_price=decimal.Decimal(20))
    position_inst.entry_price = decimal.Decimal(10)
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == decimal.Decimal(-50)
    assert position_inst.is_short()
    assert position_inst.already_reduced_size == constants.ZERO
    # ensure PNL is properly saved
    assert position_inst.realised_pnl == decimal.Decimal(1000)
    assert [
        t.realised_pnl
        for t in exchange_manager_inst.exchange_personal_data.transactions_manager.transactions.values()
        if isinstance(t, personal_data.RealisedPnlTransaction)
    ] == [decimal.Decimal(1000)]


async def test_update_size_from_order_with_long_close_position_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=constants.ONE_HUNDRED)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(2),
                      price=decimal.Decimal(20))
    limit_sell.close_position = True
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == constants.ZERO


async def test_update_size_from_order_with_long_reduce_only_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=constants.ONE_HUNDRED)
    position_inst.entry_price = decimal.Decimal(15)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=constants.ONE_HUNDRED * constants.ONE_HUNDRED,
                      price=decimal.Decimal(20))
    limit_sell.reduce_only = True
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == constants.ZERO

    # reduce only with closed position
    position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
    await position_inst.update(update_size=constants.ZERO)
    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=constants.ONE_HUNDRED * constants.ONE_HUNDRED,
                      price=decimal.Decimal(20))
    limit_sell.reduce_only = True
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == constants.ZERO


async def test_update_size_from_order_with_long_oversold_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=constants.ONE_HUNDRED)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=constants.ONE_HUNDRED ** decimal.Decimal(5),
                      price=decimal.Decimal(20))
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == decimal.Decimal("-9999999900")
    assert not position_inst.is_long()


async def test_update_size_from_order_with_short_one_way_position(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=decimal.Decimal(2),
                     price=decimal.Decimal(20))
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == decimal.Decimal(-98)


async def test_update_size_from_reversing_order_with_short_one_way_position(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=constants.ONE_HUNDRED + decimal.Decimal(50), # buying 150 => end size is +50
                     price=decimal.Decimal(10),
                     filled_price=decimal.Decimal(10))
    position_inst.entry_price = decimal.Decimal(20)
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == decimal.Decimal(50)
    assert position_inst.is_long()
    assert position_inst.already_reduced_size == constants.ZERO
    # ensure PNL is properly saved
    assert position_inst.realised_pnl == decimal.Decimal(1000)
    assert [
        t.realised_pnl
        for t in exchange_manager_inst.exchange_personal_data.transactions_manager.transactions.values()
        if isinstance(t, personal_data.RealisedPnlTransaction)
    ] == [decimal.Decimal(1000)]


async def test_update_size_from_order_with_short_close_position_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=decimal.Decimal(2),
                     price=decimal.Decimal(20))
    buy_limit.close_position = True
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == constants.ZERO


async def test_update_size_from_order_with_short_reduce_only_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)
    position_inst.entry_price = decimal.Decimal(12)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=constants.ONE_HUNDRED * constants.ONE_HUNDRED,
                     price=decimal.Decimal(20))
    buy_limit.reduce_only = True
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == constants.ZERO

    # reduce only with closed position
    position_inst = personal_data.LinearPosition(trader_inst, symbol_contract)
    await position_inst.update(update_size=constants.ZERO)
    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=constants.ONE_HUNDRED * constants.ONE_HUNDRED,
                     price=decimal.Decimal(20))
    buy_limit.reduce_only = True
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == constants.ZERO


async def test_update_size_from_order_realized_pnl_position(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)

    mark_price = decimal.Decimal(20)
    position_inst.entry_price = decimal.Decimal(10)
    await position_inst.update(mark_price=mark_price, update_size=-decimal.Decimal(50))

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=mark_price,
                     quantity=constants.ONE,
                     price=decimal.Decimal(25))

    if not os.getenv('CYTHON_IGNORE'):
        with mock.patch.object(buy_limit, "get_total_fees", mock.Mock(return_value=5)):
            await position_inst.update_from_order(buy_limit)
        assert position_inst.size == decimal.Decimal("-49")
        assert position_inst.entry_price == decimal.Decimal("20")
        assert position_inst.realised_pnl == decimal.Decimal("-5")


async def test_update_size_from_order_with_short_overbought_one_way_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=constants.ONE_HUNDRED ** decimal.Decimal(5),
                     price=decimal.Decimal(20))
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == decimal.Decimal("9999999900")
    assert not position_inst.is_short()


async def test_update_size_from_order_with_long_oversold_hedge_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=constants.ONE_HUNDRED)

    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=constants.ONE_HUNDRED ** decimal.Decimal(5),
                      price=decimal.Decimal(20))
    # cannot switch side
    await position_inst.update_from_order(limit_sell)
    assert position_inst.size == constants.ZERO
    assert position_inst.is_idle()


async def test_update_size_from_order_with_short_overbought_hedge_position(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-constants.ONE_HUNDRED)

    buy_limit = BuyLimitOrder(trader_inst)
    buy_limit.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(10),
                     quantity=constants.ONE_HUNDRED ** decimal.Decimal(5),
                     price=decimal.Decimal(20))
    # cannot switch side
    await position_inst.update_from_order(buy_limit)
    assert position_inst.size == constants.ZERO
    assert position_inst.is_idle()


async def test__calculates_size_update_from_filled_order_long_increasing_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_increasing_one_way_reduce_only(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == constants.ZERO


async def test__calculates_size_update_from_filled_order_long_decreasing_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_decreasing_reduce_only_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_switching_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(20),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(20)


async def test__calculates_size_update_from_filled_order_long_switching_reduce_only_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(20),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(10)


async def test__calculates_size_update_from_filled_order_short_increasing_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_increasing_one_way_reduce_only(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == constants.ZERO


async def test__calculates_size_update_from_filled_order_short_decreasing_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_decreasing_reduce_only_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_switching_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(30),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(30)


async def test__calculates_size_update_from_filled_order_short_switching_reduce_only_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(30),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(10)


async def test__calculates_size_update_from_filled_order_long_increasing_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_increasing_one_way_reduce_only(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == constants.ZERO


async def test__calculates_size_update_from_filled_order_long_decreasing_one_way_reduce_only(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_buy = SellMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(15),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(-10)


async def test__calculates_size_update_from_filled_order_long_decreasing_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_decreasing_reduce_only_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_long_switching_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(20),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(10)


async def test__calculates_size_update_from_filled_order_long_switching_reduce_only_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(20),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(10)


async def test__calculates_size_update_from_filled_order_short_increasing_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == -decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_increasing_one_way_reduce_only(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(10))
    market_sell.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_sell, position_inst.get_quantity_to_close()) == constants.ZERO


async def test__calculates_size_update_from_filled_order_short_decreasing_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_decreasing_reduce_only_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(5),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(5)


async def test__calculates_size_update_from_filled_order_short_switching_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(30),
                      price=decimal.Decimal(10))
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()
    ) == decimal.Decimal(10)


async def test__calculates_size_update_from_filled_order_short_switching_reduce_only_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=False)
    await position_inst.update(update_size=-decimal.Decimal(10))
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(30),
                      price=decimal.Decimal(10))
    market_buy.reduce_only = True
    if os.getenv('CYTHON_IGNORE'):
        return
    assert position_inst._calculates_size_update_from_filled_order(
        market_buy, position_inst.get_quantity_to_close()) == decimal.Decimal(10)


async def test_is_order_increasing_size_one_way(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=True)
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)

    # orders
    buy_market = BuyMarketOrder(trader_inst)
    reduce_buy_market = BuyMarketOrder(trader_inst)
    reduce_buy_market.reduce_only = True
    close_buy_market = BuyMarketOrder(trader_inst)
    close_buy_market.close_position = True
    sell_market = SellMarketOrder(trader_inst)
    reduce_sell_market = SellMarketOrder(trader_inst)
    reduce_sell_market.reduce_only = True
    close_sell_market = SellMarketOrder(trader_inst)
    close_sell_market.close_position = True

    # when idle
    assert position_inst.is_order_increasing_size(buy_market)
    assert position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)

    # when long
    await position_inst.update(update_size=decimal.Decimal(10))
    assert position_inst.is_order_increasing_size(buy_market)
    assert not position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)

    # when short
    await position_inst.update(update_size=decimal.Decimal(-100))
    assert not position_inst.is_order_increasing_size(buy_market)
    assert position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)


async def test_is_order_increasing_size_hedge(
        btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear
    symbol_contract = default_contract
    symbol_contract.set_position_mode(is_one_way=False)
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)

    # orders
    buy_market = BuyMarketOrder(trader_inst)
    reduce_buy_market = BuyMarketOrder(trader_inst)
    reduce_buy_market.reduce_only = True
    close_buy_market = BuyMarketOrder(trader_inst)
    close_buy_market.close_position = True
    sell_market = SellMarketOrder(trader_inst)
    reduce_sell_market = SellMarketOrder(trader_inst)
    reduce_sell_market.reduce_only = True
    close_sell_market = SellMarketOrder(trader_inst)
    close_sell_market.close_position = True

    # when idle on long side
    position_inst.side = enums.PositionSide.LONG
    assert position_inst.is_order_increasing_size(buy_market)
    assert not position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)

    # when idle on short side
    position_inst.side = enums.PositionSide.SHORT
    assert not position_inst.is_order_increasing_size(buy_market)
    assert position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)

    # when long
    position_inst.side = enums.PositionSide.LONG
    await position_inst.update(update_size=decimal.Decimal(10))
    assert position_inst.is_order_increasing_size(buy_market)
    assert not position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)

    # when short
    position_inst.side = enums.PositionSide.SHORT
    await position_inst.update(update_size=decimal.Decimal(-100))
    assert not position_inst.is_order_increasing_size(buy_market)
    assert position_inst.is_order_increasing_size(sell_market)
    assert not position_inst.is_order_increasing_size(reduce_buy_market)
    assert not position_inst.is_order_increasing_size(close_buy_market)
    assert not position_inst.is_order_increasing_size(reduce_sell_market)
    assert not position_inst.is_order_increasing_size(close_sell_market)


async def test__update_realized_pnl_from_size_update_long(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    tm = exchange_manager_inst.exchange_personal_data.transactions_manager
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    mark_price = constants.ONE
    position_size = decimal.Decimal(10)
    await position_inst.update(mark_price=mark_price, update_size=position_size)

    # update pnl to +50%
    await position_inst.update(mark_price=mark_price * decimal.Decimal(1.5))

    # reduce size by 20% --> 0.2
    await position_inst.update(-position_size * decimal.Decimal(0.2))  # = 0.2
    # = position_size / decimal.Decimal(2) * decimal.Decimal(0.8)
    assert position_inst.unrealized_pnl == decimal.Decimal('3.999999999999999944488848768')
    # = position_size / decimal.Decimal(2) * decimal.Decimal(0.2)
    assert position_inst.realised_pnl == decimal.Decimal('1.000000000000000055511151232')

    assert len(tm.transactions) == 1
    assert list(tm.transactions.values())[-1].symbol == DEFAULT_FUTURE_SYMBOL
    assert list(tm.transactions.values())[-1].realised_pnl == decimal.Decimal('1.000000000000000055511151232')

    position_size = decimal.Decimal(8)

    # reduce size by 50% --> 0.5
    await position_inst.update(-position_size * decimal.Decimal(0.5))
    # = position_size / decimal.Decimal(4) * decimal.Decimal(0.5)
    assert position_inst.unrealized_pnl == decimal.Decimal('1.999999999999999944488848768')
    # = position_size / decimal.Decimal(4) * decimal.Decimal(0.5) + previous realised pnl
    assert position_inst.realised_pnl == decimal.Decimal('3.000000000000000055511151232')

    assert len(tm.transactions) == 2
    assert list(tm.transactions.values())[-1].realised_pnl == decimal.Decimal('3.000000000000000055511151232') - \
           decimal.Decimal('1.000000000000000055511151232')

    position_size = decimal.Decimal(4)

    # Increase by 50% --> shouldn't increase realised pnl
    await position_inst.update(position_size * decimal.Decimal(0.5))
    assert position_inst.realised_pnl == decimal.Decimal('3.000000000000000055511151232')  # = previous realised pnl
    assert len(tm.transactions) == 2


async def test__update_realized_pnl_from_size_update_short(btc_usdt_future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_linear

    tm = exchange_manager_inst.exchange_personal_data.transactions_manager
    symbol_contract = default_contract
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)
    symbol_contract.set_position_mode(is_one_way=True)
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    mark_price = constants.ONE
    position_size = decimal.Decimal(100)
    await position_inst.update(mark_price=mark_price, update_size=-position_size)

    # update pnl to +50%
    await position_inst.update(mark_price=mark_price * decimal.Decimal(0.5))

    # reduce size by 30% --> 0.3
    await position_inst.update(position_size * decimal.Decimal(0.3))  # = 0.3
    assert position_inst.unrealized_pnl == decimal.Decimal('35.00000000000000055511151232')
    assert position_inst.realised_pnl == decimal.Decimal('14.99999999999999944488848768')

    assert len(tm.transactions) == 1
    assert list(tm.transactions.values())[-1].symbol == DEFAULT_FUTURE_SYMBOL
    assert list(tm.transactions.values())[-1].realised_pnl == decimal.Decimal('14.99999999999999944488848768')

    position_size = decimal.Decimal(70)

    # reduce size by 40% --> 0.4
    await position_inst.update(position_size * decimal.Decimal(0.4))
    assert position_inst.unrealized_pnl == decimal.Decimal('20.99999999999999977795539508')
    assert position_inst.realised_pnl == decimal.Decimal('29.00000000000000022204460492')

    assert len(tm.transactions) == 2
    assert list(tm.transactions.values())[-1].realised_pnl == decimal.Decimal('29.00000000000000022204460492') - \
           decimal.Decimal('14.99999999999999944488848768')

    position_size = decimal.Decimal(42)

    # Increase by 50% --> shouldn't increase realised pnl
    await position_inst.update(-position_size * decimal.Decimal(0.5))
    assert position_inst.realised_pnl == decimal.Decimal('29.00000000000000022204460492')  # = previous realised pnl
    assert len(tm.transactions) == 2


async def test_update_position_when_overliquidated_position_with_long_position_inverse_contract(
        btc_usdt_future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract, position_inst = btc_usdt_future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))
    tm = exchange_manager_inst.exchange_personal_data.transactions_manager

    position_inst = personal_data.InversePosition(trader_inst, default_contract)
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.initialize()

    # near all in position => size = 9999
    await position_inst.update(update_size=decimal.Decimal(9999), mark_price=decimal.Decimal(100))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('0.001')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('9.999')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    assert len(tm.transactions) == 0

    await position_inst.update(mark_price=decimal.Decimal(3))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('0.001')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('0.001')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    assert len(tm.transactions) == 1
    transaction = next(iter(tm.transactions.values()))
    assert transaction.realised_pnl == decimal.Decimal("-9.999")
    assert transaction.trigger_source is enums.PNLTransactionSource.LIQUIDATION



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import uuid

import time

import pytest

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.personal_data.transactions.transaction_factory as transaction_factory

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

TRANSACTION_SYMBOL = "BTC/USDT"
TRANSACTION_CURRENCY = "BTC"


async def test_create_blockchain_transaction(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 0
    t_id = str(uuid.uuid4())
    transaction = transaction_factory.create_blockchain_transaction(exchange_manager,
                                                                    is_deposit=True,
                                                                    currency=TRANSACTION_CURRENCY,
                                                                    blockchain_type=enums.BlockchainTypes.UNKNOWN.value,
                                                                    blockchain_transaction_id=t_id,
                                                                    blockchain_transaction_status=enums.BlockchainTransactionStatus.CONFIRMING,
                                                                    source_address="0x123456789",
                                                                    quantity=decimal.Decimal(50),
                                                                    transaction_fee=decimal.Decimal(0.1))
    assert transaction.currency == TRANSACTION_CURRENCY
    assert transaction.creation_time <= time.time()
    assert transaction.transaction_id == t_id
    assert transaction.blockchain_transaction_id == t_id
    assert transaction.blockchain_transaction_status is enums.BlockchainTransactionStatus.CONFIRMING
    assert transaction.source_address == "0x123456789"
    assert transaction.quantity == decimal.Decimal(50)
    assert transaction.transaction_fee == decimal.Decimal(0.1)
    assert transaction.is_deposit()
    assert not transaction.is_withdrawal()
    assert not transaction.is_validated()
    assert transaction.is_pending()
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1


async def test_create_realised_pnl_transaction(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    transaction = transaction_factory.create_realised_pnl_transaction(
        exchange_manager,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        side=enums.PositionSide.LONG,
        realised_pnl=decimal.Decimal(0.1),
        closed_quantity=decimal.Decimal(1),
        cumulated_closed_quantity=decimal.Decimal(2),
        first_entry_time=decimal.Decimal(3),
        average_entry_price=decimal.Decimal(4),
        average_exit_price=decimal.Decimal(5),
        order_exit_price=decimal.Decimal(6),
        leverage=decimal.Decimal(7),
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    assert transaction.currency == TRANSACTION_CURRENCY
    assert transaction.symbol == TRANSACTION_SYMBOL
    assert transaction.realised_pnl == decimal.Decimal(0.1)
    assert transaction.side is enums.PositionSide.LONG
    assert transaction.closed_quantity == decimal.Decimal(1)
    assert transaction.cumulated_closed_quantity == decimal.Decimal(2)
    assert transaction.first_entry_time == decimal.Decimal(3)
    assert transaction.average_entry_price == decimal.Decimal(4)
    assert transaction.average_exit_price == decimal.Decimal(5)
    assert transaction.order_exit_price == decimal.Decimal(6)
    assert transaction.leverage == decimal.Decimal(7)
    assert transaction.creation_time <= time.time()
    assert transaction.trigger_source is enums.PNLTransactionSource.LIMIT_ORDER
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1


async def test_create_fee_transaction(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    o_id = str(uuid.uuid4())
    transaction = transaction_factory.create_fee_transaction(exchange_manager,
                                                             currency=TRANSACTION_CURRENCY,
                                                             symbol=TRANSACTION_SYMBOL,
                                                             quantity=decimal.Decimal(10), order_id=o_id)
    assert transaction.currency == TRANSACTION_CURRENCY
    assert transaction.symbol == TRANSACTION_SYMBOL
    assert transaction.creation_time <= time.time()
    assert transaction.order_id == o_id
    assert transaction.transaction_id
    assert transaction.quantity == decimal.Decimal(10)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1


async def test_create_transfer_transaction(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    transaction = transaction_factory.create_transfer_transaction(exchange_manager,
                                                                  currency=TRANSACTION_CURRENCY,
                                                                  symbol=TRANSACTION_SYMBOL)
    assert transaction.currency == TRANSACTION_CURRENCY
    assert transaction.symbol == TRANSACTION_SYMBOL
    assert transaction.creation_time <= time.time()
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import typing

import octobot_commons.symbols as commons_symbols
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants
import octobot_trading.personal_data as personal_data


def get_portfolio(exchange_manager, as_decimal=True) -> dict:
    return format_portfolio(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio,
        as_decimal
    )


def get_portfolio_historical_values(exchange_manager, currency, time_frame, from_timestamp=None, to_timestamp=None) \
        -> list:
    if exchange_manager.exchange_personal_data.portfolio_manager:
        return exchange_manager.exchange_personal_data.portfolio_manager.get_portfolio_historical_values(
            currency, time_frame, from_timestamp, to_timestamp
        )
    return []


def get_portfolio_reference_market(exchange_manager) -> str:
    return exchange_manager.exchange_personal_data.portfolio_manager.reference_market


def get_global_portfolio_currencies_values(exchange_managers: list) -> dict:
    currencies_values = {}
    for exchange in exchange_managers:
        this_currency_values = (
            exchange.exchange_personal_data.portfolio_manager
            .portfolio_value_holder.get_current_crypto_currencies_values()
        )
        for currency, value in this_currency_values.items():
            if currency not in currencies_values:
                currencies_values[currency] = value
    return currencies_values


def get_portfolio_currency(exchange_manager, currency) -> personal_data.Asset:
    return exchange_manager.exchange_personal_data.portfolio_manager.portfolio.get_currency_portfolio(currency)


def get_origin_portfolio(exchange_manager, as_decimal=True) -> dict:
    return format_portfolio(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio.portfolio,
        as_decimal
    )


def set_simulated_portfolio_initial_config(exchange_manager, portfolio_content):
    if exchange_manager.exchange_personal_data.portfolio_manager:
        exchange_manager.exchange_personal_data.portfolio_manager.set_forced_portfolio_initial_config(
            portfolio_content
        )


def format_portfolio(
    portfolio: dict[str, typing.Union[personal_data.Asset, dict[str, typing.Union[float, decimal.Decimal]]]],
    as_decimal: bool
) -> dict[str, typing.Union[personal_data.Asset, dict[str, typing.Union[float, decimal.Decimal]]]]:
    if not portfolio:
        return portfolio
    first_value = next(iter(portfolio.values()))
    if isinstance(first_value, personal_data.Asset):
        if as_decimal:
            return portfolio
        return personal_data.portfolio_to_float(portfolio)
    return personal_data.format_dict_portfolio_values(portfolio, as_decimal)


def parse_decimal_portfolio(portfolio, as_decimal) -> dict:
    return personal_data.parse_decimal_portfolio(portfolio, as_decimal=as_decimal)


async def refresh_real_trader_portfolio(exchange_manager) -> bool:
    return await exchange_channel.get_chan(octobot_trading.constants.BALANCE_CHANNEL, exchange_manager.id). \
        get_internal_producer(). \
        refresh_real_trader_portfolio(True)


def get_draw_down(exchange_manager) -> decimal.Decimal:
    return personal_data.get_draw_down(exchange_manager)


async def get_coefficient_of_determination(exchange_manager, use_high_instead_of_end_balance=True):
    return await personal_data.get_coefficient_of_determination(exchange_manager,
                                                                use_high_instead_of_end_balance)


def get_usd_like_symbol_from_symbols(currency: str, symbols) -> str:
    return personal_data.ValueConverter.get_usd_like_symbol_from_symbols(currency, symbols)


def get_usd_like_symbols_from_symbols(currency: str, symbols) -> list:
    return personal_data.ValueConverter.get_usd_like_symbols_from_symbols(currency, symbols)


def can_convert_symbol_to_usd_like(symbol: str) -> bool:
    return personal_data.ValueConverter.can_convert_symbol_to_usd_like(symbol)


def is_usd_like_coin(coin) -> bool:
    return commons_symbols.is_usd_like_coin(coin)


def resolve_sub_portfolios(
    master_portfolio: personal_data.SubPortfolioData,
    sub_portfolios: list[personal_data.SubPortfolioData],
    allowed_filling_assets: list[str],
    forbidden_filling_assets: list[str],
    market_prices: dict[str, float],
) -> (personal_data.SubPortfolioData, list[personal_data.SubPortfolioData]):
    return personal_data.resolve_sub_portfolios(
        master_portfolio, sub_portfolios, allowed_filling_assets, forbidden_filling_assets, market_prices
    )


def get_portfolio_filled_orders_deltas(
    previous_portfolio_content: dict[str, dict[str, decimal.Decimal]],
    updated_portfolio_content: dict[str, dict[str, decimal.Decimal]],
    filled_orders: list[personal_data.Order]
) -> (dict[str, dict[str, decimal.Decimal]], dict[str, dict[str, decimal.Decimal]]):
    return personal_data.get_portfolio_filled_orders_deltas(
        previous_portfolio_content, updated_portfolio_content, filled_orders
    )


def get_assets_delta_from_orders(
    orders: list[personal_data.Order]
) -> (dict[str, decimal.Decimal], dict[str, decimal.Decimal], dict[str, decimal.Decimal]):
    return personal_data.get_assets_delta_from_orders(orders)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import typing

import octobot_commons.dataclasses
import octobot_trading.exchanges


@dataclasses.dataclass
class IncompatibleAssetDetails(
    octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass
):
    symbol: str = ""
    updated_at: float = 0


@dataclasses.dataclass
class ExchangeAuthDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    api_key: str = ""
    api_secret: str = ""
    api_password: str = ""
    access_token: str = ""
    exchange_type: str = ""
    sandboxed: bool = False
    broker_enabled: bool = False
    encrypted: str = ""
    exchange_account_id: typing.Union[str, None] = None
    incompatible_assets: typing.Union[list[IncompatibleAssetDetails], None] = dataclasses.field(default_factory=list)

    # pylint: disable=E1134
    def __post_init__(self):
        if self.incompatible_assets and isinstance(self.incompatible_assets[0], dict):
            self.incompatible_assets = (
                [IncompatibleAssetDetails.from_dict(asset) for asset in self.incompatible_assets]
                if self.incompatible_assets else []
            )


@dataclasses.dataclass
class ExchangeDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    name: str = ""


@dataclasses.dataclass
class MarketDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    id: str = ""
    symbol: str = ""
    details: octobot_trading.exchanges.SymbolDetails = \
        dataclasses.field(default_factory=octobot_trading.exchanges.SymbolDetails)
    time_frame: str = ""
    close: list[float] = dataclasses.field(default_factory=list)
    open: list[float] = dataclasses.field(default_factory=list)
    high: list[float] = dataclasses.field(default_factory=list)
    low: list[float] = dataclasses.field(default_factory=list)
    volume: list[float] = dataclasses.field(default_factory=list)
    time: list[float] = dataclasses.field(default_factory=list)

    def has_full_candles(self):
        return self.close and self.open and self.high and self.low and self.time


@dataclasses.dataclass
class OrdersDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    open_orders: list[dict] = dataclasses.field(default_factory=list)
    missing_orders: list[dict] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class PortfolioDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    initial_value: float = 0    # value of the portfolio content (not the full_content)
    content: dict = dataclasses.field(default_factory=dict) # might be a subset of global_content
    full_content: dict = dataclasses.field(default_factory=dict)  # full exchange portfolio
    asset_values: dict = dataclasses.field(default_factory=dict)    # unitary value of each asset in reference market


@dataclasses.dataclass
class PositionDetails(octobot_commons.dataclasses.FlexibleDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    position: dict = dataclasses.field(default_factory=dict)
    contract: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ExchangeData(octobot_commons.dataclasses.MinimizableDataclass, octobot_commons.dataclasses.UpdatableDataclass):
    auth_details: ExchangeAuthDetails = dataclasses.field(default_factory=ExchangeAuthDetails)
    exchange_details: ExchangeDetails = dataclasses.field(default_factory=ExchangeDetails)
    markets: list[MarketDetails] = dataclasses.field(default_factory=list)
    portfolio_details: PortfolioDetails = dataclasses.field(default_factory=PortfolioDetails)
    orders_details: OrdersDetails = dataclasses.field(default_factory=OrdersDetails)
    trades: list[dict] = dataclasses.field(default_factory=list)
    positions: list[PositionDetails] = dataclasses.field(default_factory=list)

    # pylint: disable=E1134
    def __post_init__(self):
        if self.markets and isinstance(self.markets[0], dict):
            self.markets = [MarketDetails.from_dict(market) for market in self.markets] if self.markets else []
        if self.positions and isinstance(self.positions[0], dict):
            self.positions = [PositionDetails.from_dict(position) for position in self.positions] \
                if self.positions else []

    def get_price(self, symbol):
        for market in self.markets:
            if market.symbol == symbol:
                return market.close[-1]
        raise KeyError(symbol)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import copy
import mock

import octobot_commons.constants
import octobot_commons.logging
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data
import octobot_trading.api as trading_api


def test_resolve_sub_portfolios_no_filling_assets():
    master_pf = _sub_pf(0, _content({"BTC": 0.1, "ETH": 9.9999999, "USDT": 100}))
    origin_master_pf = copy.deepcopy(master_pf)

    assert personal_data.resolve_sub_portfolios(master_pf, [], [], [], {}) == (
        origin_master_pf, []
    )

    sub_pf_btc = _sub_pf(0, _content({"BTC": 0.01}))
    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pf_btc], [], [], {}) == (
        _sub_pf(0, _content({"BTC": 0.09, "ETH": 9.9999999, "USDT": 100})),
        [_sub_pf(0, _content({"BTC": 0.01}))]
    )

    sub_pf_btc_usdt = _sub_pf(1, _content({"BTC": 0.05, "USDT": 100}))
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt], [], [], {}
    ) == (
        _sub_pf(0, _content({"BTC": 0.04, "ETH": 9.9999999, "USDT": 0})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100}))
        ]
    )

    # not enough BTC and USDT in master portfolio for this sub portfolio
    sub_pf_btc_usdt_2 = _sub_pf(2, _content({"BTC": 0.06, "USDT": 40, "TRX": 11}))
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2], [], [], {}
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 9.9999999, "USDT": 0})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                _content({"BTC": 0.04, "USDT": 0, "TRX": 0}),
                missing_funds=_missing_funds({"BTC": 0.02, "USDT": 40, "TRX": 11})
            )  # adapted to available in master
        ]
    )

    # portfolio with higher priority takes BTC before others
    priority_sub_pf_btc = _sub_pf(0.1, _content({"BTC": 0.02}))
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2, priority_sub_pf_btc], [], [], {}
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 9.9999999, "USDT": 0})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(0.1, _content({"BTC": 0.02})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                _content({"BTC": 0.02, "USDT": 0, "TRX": 0}),
                missing_funds=_missing_funds({"BTC": 0.04, "USDT": 40, "TRX": 11})
            )  # adapted to available in master
        ]
    )


def test_resolve_sub_portfolios_no_filling_assets_with_locked_funds():
    master_pf = _sub_pf(
        0,
        _content_with_available({"BTC": (0.019998, 0.1), "ETH": (0.019998, 0.1), "USDT": (0.019998, 0.1)})
    )
    sub_pf_btc = _sub_pf(0, _content({"BTC": 0.1}))
    sub_pf_btc.locked_funds_by_asset = trading_api.get_orders_locked_amounts_by_asset([
        _open_order("BTC/USDT", 0.06, 400, "sell", 0.000001, "BTC"),
        _open_order("BTC/USDT", 0.02, 600, "sell", 0.000001, "BTC"),
    ])
    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pf_btc], [], [], {}) == (
        _sub_pf(0, _content_with_available({"BTC": (0, 0), "ETH": (0.019998, 0.1), "USDT": (0.019998, 0.1)})),
        [_sub_pf(0, _content_with_available({"BTC": (0.019998, 0.1)}))]
    )

    master_pf = _sub_pf(
        0,
        _content_with_available({"BTC": (0.019997, 0.1), "ETH": (0.1, 0.1), "USDT": (0, 100)})
    )
    sub_pf_1 = _sub_pf(
        0, _content({"BTC": 0.085, "ETH": 0.1}),
        locked_funds_by_asset = trading_api.get_orders_locked_amounts_by_asset([
            _open_order("BTC/USDT", 0.06, 400, "sell", 0.000001, "BTC"),
            _open_order("BTC/USDT", 0.02, 600, "sell", 0.000001, "BTC"),
        ])
    )
    sub_pf_2 = _sub_pf(
        0, _content({"BTC": 0.015, "USDT": 100}),
        locked_funds_by_asset = trading_api.get_orders_locked_amounts_by_asset([
            _open_order("BTC/USDT", 0.1, 30, "buy", 0.000001, "BTC"),
            _open_order("BTC/USDT", 0.1, 60, "buy", 5, "USDT"),
            _open_order("ETH/USDT", 0.1, 4, "buy", 1, "USDT"),
        ])
    )
    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pf_1, sub_pf_2], [], [], {}) == (
        _sub_pf(0, _content_with_available({"BTC": (0, 0), "ETH": (0, 0), "USDT": (0, 0)})),
        [
            _sub_pf(0, _content_with_available({"BTC": (0.004998, 0.085), "ETH": (0.1, 0.1)})),
            _sub_pf(0, _content_with_available({"BTC": (0.015, 0.015), "USDT": (0, 100)}))
        ]
    )


def test_resolve_sub_portfolios_with_filling_assets():
    master_pf = _sub_pf(0, _content({"BTC": 0.1, "ETH": 9.9999999, "USDT": 100, "USDC": 100}))
    origin_master_pf = copy.deepcopy(master_pf)
    filling_assets = ["USDC", "SOL"]
    market_prices = {
        "SOL/USDT": 150,
        "BTC/USDC": 83000,
        "ETH/USDC": 2000,
        "USDT/USDC": 1,
        "USDC/TRX": 0.5,    # equivalent to TRX/USDT: 2
    }

    assert personal_data.resolve_sub_portfolios(master_pf, [], filling_assets, [], market_prices) == (
        origin_master_pf, []
    )

    sub_pf_btc = _sub_pf(0, _content({"BTC": 0.01}))
    sub_pf_btc_usdt = _sub_pf(1, _content({"BTC": 0.05, "USDT": 100}))
    # not enough BTC and USDT in master portfolio for this sub portfolio but can use USDC instead
    sub_pf_btc_usdt_2 = _sub_pf(2, _content({"BTC": 0.04001, "USDT": 40, "TRX": 11}))

    # no allowed filling assets
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2], [], [], market_prices
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 9.9999999, "USDT": 0, "USDC": 100})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                _content({"BTC": 0.04, "USDT": 0, "TRX": 0}),
                # BTC not considered as missing since the missing amount % is very low
                missing_funds=_missing_funds({"USDT": 40, "TRX": 11})
            )  # adapted to available in master
        ]
    )

    # no market_prices
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2], filling_assets, [], {}
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 9.9999999, "USDT": 0, "USDC": 100})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                _content({"BTC": 0.04, "USDT": 0, "TRX": 0}),
                # BTC not considered as missing since the missing amount % is very low
                missing_funds=_missing_funds({"USDT": 40, "TRX": 11})
            )  # adapted to available in master
        ]
    )

    # now adapt sub portfolio 3
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2], filling_assets, [], market_prices
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 9.9999999, "USDT": 0, "USDC": 38})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                # 62 = 40 (for USDT) + 22 (for 11 TRX)
                _content({"BTC": 0.04, "USDT": 0, "USDC": 62, "TRX": 0}),
                # BTC not considered as missing since the missing amount % is very low
                funds_deltas=_content({"USDC": 62}),
            )  # adapted to available in master
        ]
    )

    # now adapt sub portfolio 3 and partially 4
    sub_pf_usdt_ada = _sub_pf(5, _content({"ETH": 1, "USDT": 4, "ADA": 1199}))
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2, sub_pf_usdt_ada], filling_assets, [], market_prices
    ) == (
        _sub_pf(0, _content({"BTC": 0, "ETH": 8.9999999, "USDT": 0, "USDC": 34})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(1, _content({"BTC": 0.05, "USDT": 100})),
            _sub_pf(
                2,
                # 62 = 40 (for USDT) + 22 (for 11 TRX)
                _content({"BTC": 0.04, "USDT": 0, "USDC": 62, "TRX": 0}),
                # BTC not considered as missing since the missing amount % is very low
                funds_deltas=_content({"USDC": 62}),
            ),
            _sub_pf(
                5,
                _content({"ETH": 1, "USDT": 0, "USDC": 4, "ADA": 0}),
                funds_deltas=_content({"USDC": 4}),
                missing_funds=_missing_funds({"ADA": 1199}),
            ),
        ]
    )

    # with forbidden assets
    forbidden_assets = ["USDT", "ADA"]
    assert personal_data.resolve_sub_portfolios(
        master_pf, [sub_pf_btc, sub_pf_btc_usdt, sub_pf_btc_usdt_2, sub_pf_usdt_ada], filling_assets, forbidden_assets, market_prices
    ) == (
        # USDT is kept in master portfolio
        _sub_pf(0, _content({"BTC": 0, "ETH": 8.9999999, "USDT": 100, "USDC": 0})),
        [
            _sub_pf(0, _content({"BTC": 0.01})),
            _sub_pf(
                1,
                _content({"BTC": 0.05,"USDC": 100,"USDT": 0}),
                funds_deltas=_content({"USDC": 100})
            ),
            _sub_pf(
                2,
                # 62 = 40 (for USDT) + 22 (for 11 TRX)
                _content({"BTC": 0.04, "USDC": 0, "USDT": 0, "TRX": 0}),
                # BTC not considered as missing since the missing amount % is very low
                funds_deltas={},
                missing_funds=_missing_funds({"TRX": 11, "USDT": 40}),
            ),
            _sub_pf(
                5,
                _content({"ADA": 0, "ETH": 1, "USDC": 0, "USDT": 0}),
                funds_deltas={},
                missing_funds=_missing_funds({"ADA": 1199, "USDT": 4}),
            ),
        ]
    )


def test_resolve_sub_portfolios_with_unexpected_deltas():
    # more delta in post trade portfolio than actual orders => do not result in negative subportfolio
    master_pf = _sub_pf(0, _content({"BTC": 0.05, "TRX": 0, "USDT": 200}))
    sub_pft = _sub_pf(
        0,
        _content({"BTC": 0.05, "TRX": 13.1, "USDT": 100}),
        funds_deltas=_content({"BTC": -0.05, "TRX": -13.2, "USDT": 100}),
    )

    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pft], [], [], {}) == (
        _sub_pf(0,  _content({"BTC": 0.05, "TRX": 0, "USDT": 0})),
        [_sub_pf(0, _content({"BTC": 0, "TRX": 0, "USDT": 200}))]
    )


def test_get_portfolio_filled_orders_deltas():
    pre_trade_content = _content({"BTC": 0.1, "USDT": 1000})
    post_trade_content = _content({"BTC": 0.2, "USDT": 500})
    error_log = mock.Mock()
    with mock.patch.object(octobot_commons.logging, "get_logger", mock.Mock(return_value=mock.Mock(error=error_log))):
        # no filled order
        assert personal_data.get_portfolio_filled_orders_deltas(pre_trade_content, post_trade_content, []) == (
            {}, {}
        )
        error_log.assert_not_called()

        filled_orders = [
            _order("BTC/USDT", 0.06, 100, "buy"),
            _order("BTC/USDT", 0.05, 451, "buy"),   # 1 will be ignored, (allowed error margin)
            _order("BTC/USDT", 0.01, 50, "sell"),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"BTC": 0.1, "USDT": -500}),   # all orders found in deltas
            {}
        )
        error_log.assert_not_called()

        filled_orders = [
            _order("BTC/USDT", 0.06, 100, "buy"),
            _order("BTC/USDT", 0.05, 450, "buy"),
            _order("BTC/USDT", 0.01, 50, "buy"),    # not found in portfolio delta
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            {},
            _content({"BTC": 0.1, "USDT": -500}),   # all missing: orders can't explain this delta
        )
        error_log.assert_not_called()

        filled_orders = [
            _order("SOL/USDT", 1, 200, "buy"),  # won't be found
            _order("BTC/USDT", 0.06, 100.01, "buy"),    # small change compared to portfolio
            _order("BTC/USDT", 0.01, 50, "buy"),    # not found in portfolio delta
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"BTC": 0.07, "USDT": -350.01}),   # adapted to orders explained orders delta
            _content({"SOL": 1}),   # not found in portfolio delta
        )
        error_log.assert_not_called()

        post_trade_content = _content({"USDT": 500, "SOL": 1})  # now has SOL but no BTC
        filled_orders = [
            _order("SOL/USDT", 1, 200, "buy"),  # won't be found
            _order("BTC/USDT", 0.06, 100.01, "buy"),    # small change compared to portfolio
            _order("BTC/USDT", 0.01, 50, "buy"),    # not found in portfolio delta
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"SOL": 1, "USDT": -350.01}),   # adapted to orders explained orders delta
            _content({"BTC": -0.1}),   # not found in portfolio delta
        )
        error_log.assert_not_called()

        # only sell orders
        post_trade_content = _content({"USDT": 2000})  # now has SOL but no BTC
        filled_orders = [
            _order("BTC/USDT", 0.06, 400, "sell"),    # small change compared to portfolio
            _order("BTC/USDT", 0.04, 600, "sell"),    # not found in portfolio delta
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"BTC": -0.1, "USDT": 1000}),
            {},
        )
        error_log.assert_not_called()

        # only equivalent sell and buy orders: no BTC delta
        pre_trade_content = _content({"BTC": 0.1, "USDT": 1000})
        post_trade_content = _content({"BTC": 0.1, "USDT": 995})  # now has SOL but no BTC
        filled_orders = [
            _order("BTC/USDT", 0.05, 550, "sell"),
            _order("BTC/USDT", 0.05, 555, "buy"),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"USDT": -5}),
            {}, # BTC is not a missing delta as there is no delta (delta = 0)
        )
        error_log.assert_not_called()


def test_get_portfolio_filled_orders_deltas_considering_fees():
    pre_trade_content = _content({"BTC": 0.1, "USDT": 600.2})
    post_trade_content = _content({"BTC": 0.2, "USDT": 50.2})    # paid 50 USDT in fees
    error_log = mock.Mock()
    with mock.patch.object(octobot_commons.logging, "get_logger", mock.Mock(return_value=mock.Mock(error=error_log))):
        # no filled order
        assert personal_data.get_portfolio_filled_orders_deltas(pre_trade_content, post_trade_content, []) == (
            {}, {}
        )
        error_log.assert_not_called()

        # with fees: delta is found
        filled_orders = [
            _order("BTC/USDT", 0.06, 100, "buy", {"USDT": 25}),
            _order("BTC/USDT", 0.05, 450, "buy", {"USDT": 25}),
            _order("BTC/USDT", 0.01, 50, "sell", {"BTC": 0.0000001}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # all orders found in deltas because fees have been taken into account
            _content({"BTC": 0.1, "USDT": -550}),
            {}
        )
        error_log.assert_not_called()

        # without fees explanation
        filled_orders = [
            _order("BTC/USDT", 0.06, 100, "buy", {"USDT": 0.025}),
            _order("BTC/USDT", 0.05, 450, "buy", {"USDT": 0.025}),
            _order("BTC/USDT", 0.01, 50, "sell", {"BTC": 0.0000001}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # fees can't explain the additional -50 only consider -500
            _content({"BTC": 0.1, "USDT": -500}),
            {}
        )
        error_log.assert_not_called()

        # with small numbers
        pre_trade_content = _content({"BTC": 0.1, "USDT": 0.2})
        post_trade_content = _content({"BTC": 0.089, "USDT": 0.1})    # paid 0.1 USDT and 0.001 BTC in fees
        filled_orders = [
            _order("BTC/USDT", 0.1, 100, "sell", {"USDT": 0.1}),
            _order("BTC/USDT", 0.09, 100, "buy", {"BTC": 0.001}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"BTC": -0.011, "USDT": -0.1}),
            {}
        )
        error_log.assert_not_called()

        # with USDT holdings increasing

        # A. return ignored assets because no explaining fees
        pre_trade_content = _content({"BTC": 0.1, "USDT": 0.2})
        post_trade_content = _content({"BTC": 0.089, "USDT": 0.8})
        filled_orders = [
            _order("BTC/USDT", 0.1, 100, "sell", {"USDT": 0.000001}),
            _order("BTC/USDT", 0.09, 99.3, "buy", {"BTC": 0.000000001}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # 0.6 is NOT explained by 0.7 from order delta - 0.000001 fees
            _content({"BTC": -0.01}),   # takes a bit too much of post trade portfolio (fees are not considered),
            # will be adapted later on if necessary
            _content({"USDT": 0.6}) # ignored assets
        )
        error_log.assert_not_called()

        # B. success with explaining fees
        pre_trade_content = _content({"BTC": 0.1, "USDT": 0.2})
        post_trade_content = _content({"BTC": 0.089, "USDT": 0.8})
        filled_orders = [
            _order("BTC/USDT", 0.1, 100, "sell", {"USDT": 0.1}),
            _order("BTC/USDT", 0.09, 99.3, "buy", {"BTC": 0.001}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # 0.6 is explained by 0.7 from order delta - 0.1 fees
            _content({"BTC": -0.011, "USDT": 0.6}),
            {}
        )
        error_log.assert_not_called()

        # only equivalent sell and buy orders: no BTC delta
        pre_trade_content = _content({"BTC": 0.1, "USDT": 1000})
        post_trade_content = _content({"BTC": 0.1, "USDT": 995})
        filled_orders = [
            _order("BTC/USDT", 0.05, 550, "sell", {"USDT": 1}),
            _order("BTC/USDT", 0.05, 553, "buy", {"USDT": 1}),
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            _content({"USDT": -5}),
            {}, # BTC is not a missing delta as there is no delta (delta = 0)
        )
        error_log.assert_not_called()

        # actual fee currency is different from expected fee currency, delta is explained by these fees
        BTC_fees = 0.00000109889 # 10% of computed BTC delta is fees but it's only fetched from portfolio, not expected order fees
        pre_trade_content = _content({"BTC": 0.00226262681, "USDT": 4376.532183487584})
        post_trade_content = _content({"BTC": 0.00227318792 - BTC_fees, "USDT": 4375.22614367084})
        filled_orders = [
            _order("BTC/USDT", 0.00108723, 111.778333746, "sell", {"USDT": 0.111778333746}),
            _order("BTC/USDT", 0.00109889, 112.972595229, "buy", {"USDT": 0.111778333746}), # expects USDT as fees but was actually BTC
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # BTC delta from unexpected fees are taken into account and don't create missing deltas
            _content({"BTC": 0.00000946222, "USDT": -1.306039816744}),
            {},
        )
        error_log.assert_not_called()

        # actual fee currency is different from expected fee currency, delta is explained by these fees
        pre_trade_content = _content({"BTC": 0.00226262681, "USDT": 4376.532183487584})
        post_trade_content = _content({"BTC": 0.00227318792, "USDT": 4375.22614367084})
        filled_orders = [
            _order("BTC/USDT", 0.00108723, 111.778333746, "sell", {"USDT": 0.111778333746}), # expects USDT as fees but was actually something else (like BNB)
            _order("BTC/USDT", 0.00109779111, 112.972595229, "buy", {"USDT": 0.111778333746}), # expects USDT as fees but was actually something else (like BNB)
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # USDT delta == 4375.22614367084 - 4376.532183487584
            _content({"USDT": -1.306039816744, "BTC": 0.00001056111}),
            {},
        )
        error_log.assert_not_called()

        actual_fees = 0.2775394125  # 0.6% fees
        # actual fees is much smaller than expected, portfolio delta is accepted
        pre_trade_content = _content({"MANA": 103.42087452540375, "USDC": actual_fees})
        post_trade_content = _content({"MANA": 72.65087452540375, "USDC": 0.0827944124999931})
        filled_orders = [
            _order("MANA/USDC", 103.05, 37.005255, "sell", {"USDC": actual_fees * 2}), # expects 1.2% fees
            _order("MANA/USDC", 72.28, 36.644921175, "buy", {"USDC": actual_fees * 2}),    # expects 1.2% fees
        ]
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (
            # USDC delta from portfolio is accepted and not considered as missing
            _content({"MANA": -30.77, "USDC": -0.1947450000000069}),
            {},
        )
        error_log.assert_not_called()


def test_get_portfolio_filled_orders_deltas_considering_different_fee_tiers():
    error_log = mock.Mock()
    with mock.patch.object(octobot_commons.logging, "get_logger", mock.Mock(return_value=mock.Mock(error=error_log))):
        # base 1.2% fees
        pre_trade_usdc = 2.203362439822
        pre_trade_content = _content({"ETH": 0.01145892, "USDC": pre_trade_usdc, 'BTC': 0.01937857})
        filled_orders = [
            _order("BTC/USDC", 0.00091181, 99.2684993932, "sell", {"USDC": 1.1912219927184}),
            _order("XRP/USDC", 3.840399, 9.1942992459, "buy", {"USDC": 0.1103315909508}),
            _order("XLM/USDC", 33.59950349, 9.79704799822749, "buy", {"USDC": 0.11756457597873}),
            _order("SUI/USDC", 2.4, 9.35448, "buy", {"USDC": 0.11225376}),
            _order("SOL/USDC", 0.05705159, 9.78999848, "buy", {"USDC": 0.11747998176}),
            _order("LINK/USDC", 0.61, 9.73987, "buy", {"USDC": 0.11687844}),
            _order("ETH/USDC", 0.00387595, 9.7899907885, "buy", {"USDC": 0.117479889462}),
            _order("DOGE/USDC", 42.3, 9.778068, "buy", {"USDC": 0.117336816}),
            _order("BTC/USDC", 0.00008992, 9.7895661216, "buy", {"USDC": 0.1174747934592}),
            _order("AVAX/USDC", 0.42769768, 9.7899998952, "buy", {"USDC": 0.1174799987424}),
            _order("ADA/USDC", 12.75318927, 9.796999997214, "buy", {"USDC": 0.117563999966568}),
        ]
        max_fees_deltas = {
            "XRP": 3.840399, "SUI": 2.4, "XLM": 33.59950349, "SOL": 0.05705159, "LINK": 0.61, "DOGE": 42.3,
            "AVAX": 0.42769768, "ADA": 12.75318927, "ETH": 0.00387595, "USDC": 0.095113027520412,
            "BTC": -0.00082189
        }
        post_trade_content = _content({
            "XRP": 3.840399, "SUI": 2.4, "XLM": 33.59950349, "SOL": 0.05705159, "LINK": 0.61, "DOGE": 42.3,
            "AVAX": 0.42769768, "ADA": 12.75318927, "ETH": 0.01533487, "USDC": 2.298475467342412, "BTC": 0.01855668
        })
        assert personal_data.get_portfolio_filled_orders_deltas(
            pre_trade_content, post_trade_content, filled_orders
        ) == (_content(max_fees_deltas), {})
        error_log.assert_not_called()

        # use coinbase fees tiers for this test
        expected_fees_percent = 1.2
        for real_fee_percent in (1.2, 0.75, 0.4, 0.25, 0.20, 0.16, 0.14, 0.10, 0.08, 0.05):
            total_real_paid_fees = [decimal.Decimal(0)]

            def _get_expected_fees(cost):
                real_paid_fees = decimal.Decimal(str(cost)) * decimal.Decimal(str(real_fee_percent)) / decimal.Decimal(100)
                expected_paid_fees = decimal.Decimal(str(cost)) * decimal.Decimal(str(expected_fees_percent)) / decimal.Decimal(100)
                total_real_paid_fees[0] = total_real_paid_fees[0] + real_paid_fees
                # expected fees are always 1.2%
                return float(expected_paid_fees)

            filled_orders = [
                _order("BTC/USDC", 0.00091181, 99.2684993932, "sell", {"USDC": _get_expected_fees(99.2684993932)}),
                _order("XRP/USDC", 3.840399, 9.1942992459, "buy", {"USDC": _get_expected_fees(9.1942992459)}),
                _order("XLM/USDC", 33.59950349, 9.79704799822749, "buy", {"USDC": _get_expected_fees(9.79704799822749)}),
                _order("SUI/USDC", 2.4, 9.35448, "buy", {"USDC": _get_expected_fees(9.35448)}),
                _order("SOL/USDC", 0.05705159, 9.78999848, "buy", {"USDC": _get_expected_fees(9.78999848)}),
                _order("LINK/USDC", 0.61, 9.73987, "buy", {"USDC": _get_expected_fees(9.73987)}),
                _order("ETH/USDC", 0.00387595, 9.7899907885, "buy", {"USDC": _get_expected_fees(9.7899907885)}),
                _order("DOGE/USDC", 42.3, 9.778068, "buy", {"USDC": _get_expected_fees(9.778068)}),
                _order("BTC/USDC", 0.00008992, 9.7895661216, "buy", {"USDC": _get_expected_fees(9.7895661216)}),
                _order("AVAX/USDC", 0.42769768, 9.7899998952, "buy", {"USDC": _get_expected_fees(9.7899998952)}),
                _order("ADA/USDC", 12.75318927, 9.796999997214, "buy", {"USDC": _get_expected_fees(9.796999997214)}),
            ]

            post_trade_usdc = (
                decimal.Decimal(str(pre_trade_usdc))
                + sum(decimal.Decimal(str(o.total_cost)) for o in filled_orders if o.side == enums.TradeOrderSide.SELL)
                - sum(decimal.Decimal(str(o.total_cost)) for o in filled_orders if o.side == enums.TradeOrderSide.BUY)
                - total_real_paid_fees[0]
            )
            post_trade_content = _content({
                "XRP": 3.840399, "SUI": 2.4, "XLM": 33.59950349, "SOL": 0.05705159, "LINK": 0.61, "DOGE": 42.3,
                "AVAX": 0.42769768, "ADA": 12.75318927, "ETH": 0.01533487, "USDC": float(post_trade_usdc), "BTC": 0.01855668
            })
            local_fees_deltas = {
                "XRP": 3.840399, "SUI": 2.4, "XLM": 33.59950349, "SOL": 0.05705159, "LINK": 0.61, "DOGE": 42.3,
                "AVAX": 0.42769768, "ADA": 12.75318927, "ETH": 0.00387595, "BTC": -0.00082189,
                "USDC": float(post_trade_usdc - decimal.Decimal(str(pre_trade_usdc))),
            }

            if real_fee_percent == 1.2:
                # same result as previous (non looping) test
                assert personal_data.get_portfolio_filled_orders_deltas(
                    pre_trade_content, post_trade_content, filled_orders
                ) == (_content(max_fees_deltas), {})
            else:
                deltas, missing_deltas = personal_data.get_portfolio_filled_orders_deltas(
                    pre_trade_content, post_trade_content, filled_orders
                )
                # fix rounding issues
                local_fees_deltas["USDC"] = float(round(decimal.Decimal(local_fees_deltas["USDC"]), 8))
                assert missing_deltas == {}, f"{real_fee_percent=} {missing_deltas=}"
                for key in (octobot_commons.constants.PORTFOLIO_AVAILABLE, octobot_commons.constants.PORTFOLIO_TOTAL):
                    deltas["USDC"][key] = round(deltas["USDC"][key], 8)
                assert deltas == _content(local_fees_deltas), f"{real_fee_percent=}"

            error_log.assert_not_called()


def test_resolve_sub_portfolios_with_filling_assets_with_locked_funds():
    master_pf = _sub_pf(
        0,
        _content_with_available({
            "BTC": (0.019998, 0.1), "ETH": (0.019998, 0.1), "USDT": (10.019998, 10.1), "SOL": (0.5, 1), "USDC": (2, 2)
        })
    )
    filling_assets = ["USDC", "SOL"]
    market_prices = {
        "SOL/USDT": 150,
        "BTC/USDC": 83000,
        "ETH/USDC": 2000,
        "USDT/USDC": 1,
    }
    sub_pf_btc = _sub_pf(
        0, _content({"BTC": 0.1}),
        locked_funds_by_asset = trading_api.get_orders_locked_amounts_by_asset([
            _open_order("BTC/USDT", 0.06, 400, "sell", 0.000001, "BTC"),
            _open_order("BTC/USDT", 0.02, 600, "sell", 0.000001, "BTC"),
        ])
    )
    # nothing to fill
    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pf_btc], filling_assets, [], market_prices) == (
        _sub_pf(0, _content_with_available({
            "BTC": (0, 0), "ETH": (0.019998, 0.1), "USDT": (10.019998, 10.1), "SOL": (0.5, 1), "USDC": (2, 2)
        })),
        [_sub_pf(0, _content_with_available({"BTC": (0.019998, 0.1)}))]
    )

    master_pf = _sub_pf(
        0,
        _content_with_available({
            "BTC": (0.019998, 0.1), "ETH": (0.019998, 0.1), "USDT": (10.019998, 10.1), "SOL": (0.5, 1), "USDC": (2, 2)
        })
    )
    sub_pf_missing_usdt = _sub_pf(
        1, _content_with_available({"USDT": (140, 150), "SOL": (0, 0.5)}),
        locked_funds_by_asset = trading_api.get_orders_locked_amounts_by_asset([
            _open_order("BTC/USDT", 0.01, 10, "buy", 0.000001, "BTC"),
            _open_order("SOL/USDT", 0.5, 1, "sell", 0.000001, "USDT"),
        ])
    )
    # fill with locked amount
    assert personal_data.resolve_sub_portfolios(master_pf, [sub_pf_btc, sub_pf_missing_usdt], filling_assets, [], market_prices) == (
        _sub_pf(0, _content_with_available({
            "BTC": (0, 0), "ETH": (0.019998, 0.1), "USDT": (0, 0), "SOL": (0, 0), "USDC": (0, 0)
        })),
        [
            _sub_pf(0, _content_with_available({"BTC": (0.019998, 0.1)})),
            _sub_pf(
                1,
                _content_with_available({"USDT": (10.019998, 10.1), "SOL": (0.5, 1), "USDC": (2, 2)}),
                funds_deltas=_content_with_available({
                    "SOL": (0.5, 0.5), "USDC": (2, 2)
                }),
                missing_funds=_missing_funds({
                    "USDT": float(
                        decimal.Decimal(140)
                        - decimal.Decimal("10.019998")
                        - decimal.Decimal(str(market_prices["SOL/USDT"])) / decimal.Decimal("2")
                        - decimal.Decimal(market_prices["USDT/USDC"] * 2)
                    )
                })
            )
        ]
    )



def _sub_pf(
    priority_key: float,
    content: dict[str, dict[str, decimal.Decimal]],
    funds_deltas: dict[str, dict[str, decimal.Decimal]] = None,
    missing_funds: dict[str, decimal.Decimal] = None,
    locked_funds_by_asset: dict[str, decimal.Decimal] = None,
) -> personal_data.SubPortfolioData:
    return personal_data.SubPortfolioData(
        "", "", priority_key, content, "", funds_deltas or {}, missing_funds or {}, locked_funds_by_asset or {}
    )

def _content(content: dict[str, float]) -> dict[str, dict[str, decimal.Decimal]]:
    return {
        key: {
            octobot_commons.constants.PORTFOLIO_TOTAL: decimal.Decimal(str(val)),
            octobot_commons.constants.PORTFOLIO_AVAILABLE: decimal.Decimal(str(val)),
        }
        for key, val in content.items()
    }

def _content_with_available(content: dict[str, (float, float)]) -> dict[str, dict[str, decimal.Decimal]]:
    return {
        key: {
            octobot_commons.constants.PORTFOLIO_TOTAL: decimal.Decimal(str(total)),
            octobot_commons.constants.PORTFOLIO_AVAILABLE: decimal.Decimal(str(available)),
        }
        for key, (available, total) in content.items()
    }

def _missing_funds(funds: dict[str, float]) -> dict[str, decimal.Decimal]:
    return {
        key: decimal.Decimal(str(val))
        for key, val in funds.items()
    }

def _order(symbol: str, quantity: float, cost: float, side: str, fee: dict = None) -> personal_data.Order:
    trader = mock.Mock(exchange_manager=mock.Mock())
    order = personal_data.Order(trader)
    order.symbol = symbol
    order.origin_quantity = decimal.Decimal(str(quantity))
    order.total_cost = decimal.Decimal(str(cost))
    order.origin_price = order.total_cost / order.origin_quantity
    order.side = enums.TradeOrderSide(side)
    order.get_computed_fee = mock.Mock(return_value={
        enums.FeePropertyColumns.COST.value: decimal.Decimal(str(next(iter(fee.values())) if fee else 0)),
        enums.FeePropertyColumns.CURRENCY.value: next(iter(fee.keys())) if fee else None,
        enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: False,
    })
    return order

def _open_order(symbol: str, quantity: float, cost: float, side: str, fee_cost: float, fee_unit: str) -> personal_data.Order:
    order = _order(symbol, quantity, cost, side)
    order.is_filled = mock.Mock(return_value=False)
    order.get_computed_fee = mock.Mock(
        return_value={
            enums.FeePropertyColumns.CURRENCY.value: fee_unit,
            enums.FeePropertyColumns.COST.value: decimal.Decimal(str(fee_cost)),
        }
    )
    return order

def _locked_amounts_by_asset(amount_by_asset: dict[str, float]) -> dict[str, decimal.Decimal]:
    return {
        asset: decimal.Decimal(str(amount))
        for asset, amount in amount_by_asset.items()
    }
